/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

static void 
Date_copyToProperties(kite_thread_t *thd, kite_object_t *obj, struct tm *t)
{
    kite_set_property(thd, obj, "sec", kite_new_integer(thd, t->tm_sec));
    kite_set_property(thd, obj, "min", kite_new_integer(thd, t->tm_min));
    kite_set_property(thd, obj, "hour", kite_new_integer(thd, t->tm_hour));
    kite_set_property(thd, obj, "day_of_month", 
        kite_new_integer(thd, t->tm_mday));
    kite_set_property(thd, obj, "month", kite_new_integer(thd, t->tm_mon));
    kite_set_property(thd, obj, "year", kite_new_integer(thd, t->tm_year));
    kite_set_property(thd, obj, "day_of_week", 
        kite_new_integer(thd, t->tm_wday));
    kite_set_property(thd, obj, "day_of_year", 
        kite_new_integer(thd, t->tm_yday));
    kite_set_property(thd, obj, "dst", kite_new_boolean(thd, t->tm_isdst));
}

static void Date_copyFromProperties(kite_object_t *obj, struct tm *t)
{
    t->tm_sec = KITE_GET_INTEGER(kite_get_property(obj, "sec"));
    t->tm_min = KITE_GET_INTEGER(kite_get_property(obj, "min"));
    t->tm_hour = KITE_GET_INTEGER(kite_get_property(obj, "hour"));
    t->tm_mday = KITE_GET_INTEGER(kite_get_property(obj, "day_of_month"));
    t->tm_mon = KITE_GET_INTEGER(kite_get_property(obj, "month"));
    t->tm_year = KITE_GET_INTEGER(kite_get_property(obj, "year"));
    t->tm_wday = KITE_GET_INTEGER(kite_get_property(obj, "day_of_week"));
    t->tm_yday = KITE_GET_INTEGER(kite_get_property(obj, "day_of_year"));
    t->tm_isdst = KITE_GET_INTEGER(kite_get_property(obj, "dst"));
}

/*****************************************************************************
 * Current time in local time zone
 ****************************************************************************/
KITE_CLASS_METHOD(Date_nowLocal) {
    kite_object_t *newObject;
    
    newObject = kite_dereference_and_load(thd, "System.date");
    newObject = kite_new_instance_with_constructor(
        thd, newObject, kite_new_list(thd));
    
    time_t t = time(NULL);
    struct tm *current = localtime(&t);
    Date_copyToProperties(thd, newObject, current);
    kite_vm_push(thd, newObject);
    kite_dereference_object(newObject);
}

/*****************************************************************************
 * Current time in GMT
 ****************************************************************************/
KITE_CLASS_METHOD(Date_nowGmt) {
    kite_object_t *newObject;
    
    newObject = kite_dereference_and_load(thd, "System.date");
    newObject = kite_new_instance_with_constructor(
        thd, newObject, kite_new_list(thd));
    
    time_t t = time(NULL);
    struct tm *current = gmtime(&t);
    Date_copyToProperties(thd, newObject, current);
    kite_vm_push(thd, newObject);
    kite_dereference_object(newObject);
}

/*****************************************************************************
 * Construct date object
 ****************************************************************************/
KITE_CLASS_METHOD(Date_construct)
{
    struct tm date_values;
    kite_object_t *arg;
    int i;
    memset(&date_values, '0', sizeof(struct tm));
    
    KITE_GET_METHOD_ARGUMENT(arg, 1);
    if (arg == NULL) {
        // empty constructor
        Date_copyToProperties(thd, this, &date_values);
        return;
    }

    // all time_t struct properties are ints.
    int *cur = (int*)&date_values;
    for (i = 1; i <= 9; i++) {
        if (!KITE_IS_TYPE(arg, OBJ_INTEGER)) {
            kite_object_t *exc;
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                    "Operand must be integer");
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            return;
        }
        *cur = KITE_GET_INTEGER(arg);
        cur++;
    }
    Date_copyToProperties(thd, this, &date_values);
}

/*****************************************************************************
 * Return string representation of date
 ****************************************************************************/
KITE_CLASS_METHOD(Date_str)
{
    struct tm thistime;
    
    Date_copyFromProperties(this, &thistime);
    time_t t = mktime(&thistime);
    kite_object_t *ret = kite_new_string(thd, ctime(&t));
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Return timestamp
 ****************************************************************************/
KITE_CLASS_METHOD(Date_timestamp)
{
    struct tm thistime;
    
    Date_copyFromProperties(this, &thistime);
    kite_object_t *ret = kite_new_integer(thd, mktime(&thistime));
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Format date using given format string
 ****************************************************************************/
KITE_CLASS_METHOD(Date_format)
{
    struct tm thistime;
    kite_object_t *arg;
    char buf[2048];
    
    KITE_GET_METHOD_ARGUMENT(arg, 1);
    if (!KITE_IS_TYPE(arg, OBJ_STRING)) {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                "Operand must be string.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    Date_copyFromProperties(this, &thistime);
    strftime(buf, 2048, KITE_GET_STRING_VALUE(arg), &thistime);
    
    kite_object_t *ret = kite_new_string(thd, buf);
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Parse date using given format string
 ****************************************************************************/
KITE_CLASS_METHOD(Date_parse)
{
    struct tm thistime;
    kite_object_t *date, *format;
    
    KITE_GET_METHOD_ARGUMENT(date, 1);
    KITE_GET_METHOD_ARGUMENT(format, 2);
    if (!KITE_IS_TYPE(date, OBJ_STRING) ||
        !KITE_IS_TYPE(format, OBJ_STRING)) {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch",
                                "Operand must be string.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    kite_object_t *newObject;
    
    newObject = kite_dereference_and_load(thd, "System.date");
    newObject = kite_new_instance_with_constructor(
        thd, newObject, kite_new_list(thd));
    
    struct tm current;
    strptime(KITE_GET_STRING_VALUE(date), 
             KITE_GET_STRING_VALUE(format), &current);
    Date_copyToProperties(thd, newObject, &current);
    kite_vm_push(thd, newObject);
    kite_dereference_object(newObject);
}

KITE_MODULE_INITIALIZER(date) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *newclass = kite_new_class(thread, NULL, "date");
    
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method(thread, FUNC_COMPILED, Date_construct, 0));
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method(thread, FUNC_COMPILED, Date_construct, 9));

    kite_add_method(thread, newclass, "now_local", 
        kite_new_method(thread, FUNC_COMPILED, Date_nowLocal, 0));
    kite_add_method(thread, newclass, "now_gmt", 
        kite_new_method(thread, FUNC_COMPILED, Date_nowGmt, 0));
    kite_add_method(thread, newclass, "str", 
        kite_new_method(thread, FUNC_COMPILED, Date_str, 0));
    kite_add_method(thread, newclass, "parse", 
        kite_new_method(thread, FUNC_COMPILED, Date_parse, 1));
    kite_add_method(thread, newclass, "format", 
        kite_new_method(thread, FUNC_COMPILED, Date_format, 2));
    kite_add_method(thread, newclass, "timestamp", 
        kite_new_method(thread, FUNC_COMPILED, Date_timestamp, 0));

    /* struct tm properties */
    kite_add_property(thread, newclass, "sec", 0);
    kite_add_property(thread, newclass, "min", 0);
    kite_add_property(thread, newclass, "hour", 0);
    kite_add_property(thread, newclass, "day_of_month", 0);
    kite_add_property(thread, newclass, "month", 0);
    kite_add_property(thread, newclass, "year", 0);
    kite_add_property(thread, newclass, "day_of_week", 0);
    kite_add_property(thread, newclass, "day_of_year", 0);
    kite_add_property(thread, newclass, "dst", 0);
    
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "date", 1);
    kite_set_property(thread, parent, "date", newclass);
}
