/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include "kite_object.h"
#include "vm/kite_vm.h"

/* Common code for the integer operations */
#define PERFORM_INTEGER_OPERATION(op) \
    int result = this->builtin_data.intvalue op rhs->builtin_data.intvalue; \
    kite_vm_return(thd, kite_new_integer(thd, result));

#define PERFORM_INTEGER_BOOLEAN_OPERATION(op) \
    int result = this->builtin_data.intvalue op rhs->builtin_data.intvalue; \
    kite_vm_return(thd, kite_new_boolean(thd, result));
    
#define TYPECHECK_OPERANDS \
    kite_object_t *rhs; \
    if (args->type == OBJ_LIST) { \
        rhs = args->builtin_data.listvalue.car; \
    } else { \
        rhs = args; \
    } \
    kite_object_t *exc; \
    if (this->type != rhs->type) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Both operands must be of the same type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define CHECK_DIVIDE_BY_ZERO \
    if (rhs->builtin_data.intvalue == 0) { \
        exc = kite_new_exception(thd, "System.exceptions.DivideByZero", \
                            "Division by zero is not possible."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }
    
#define INTEGER_OPERATION(op) \
    { \
        TYPECHECK_OPERANDS \
        PERFORM_INTEGER_OPERATION(op) \
    }

#define INTEGER_BOOLEAN_OPERATION(op) \
    { \
        TYPECHECK_OPERANDS \
        PERFORM_INTEGER_BOOLEAN_OPERATION(op) \
    }
    
#define INTEGER_DIVISION_OPERATION(op) \
    { \
        TYPECHECK_OPERANDS \
        CHECK_DIVIDE_BY_ZERO \
        PERFORM_INTEGER_OPERATION(op) \
    }
    
KITE_CLASS_METHOD(Integer_add) {
    INTEGER_OPERATION(+);
}

KITE_CLASS_METHOD(Integer_subtract) {
    INTEGER_OPERATION(-);
}

KITE_CLASS_METHOD(Integer_multiply) {
    INTEGER_OPERATION(*);
}

KITE_CLASS_METHOD(Integer_divide) {
    INTEGER_DIVISION_OPERATION(/);
}

KITE_CLASS_METHOD(Integer_modulus) {
    INTEGER_DIVISION_OPERATION(%);
}

KITE_CLASS_METHOD(Integer_leftshift) {
    INTEGER_OPERATION(<<);
}

KITE_CLASS_METHOD(Integer_rightshift) {
    INTEGER_OPERATION(>>);
}

KITE_CLASS_METHOD(Integer_and) {
    INTEGER_OPERATION(&);
}

KITE_CLASS_METHOD(Integer_or) {
    INTEGER_OPERATION(|);
}

KITE_CLASS_METHOD(Integer_xor) {
    INTEGER_OPERATION(^);
}

KITE_CLASS_METHOD(Integer_lessthan) {
    INTEGER_BOOLEAN_OPERATION(<);
}

KITE_CLASS_METHOD(Integer_lessequals) {
    INTEGER_BOOLEAN_OPERATION(<=);
}

KITE_CLASS_METHOD(Integer_equals) {
    INTEGER_BOOLEAN_OPERATION(==);
}

KITE_CLASS_METHOD(Integer_greaterequals) {
    INTEGER_BOOLEAN_OPERATION(>=);
}

KITE_CLASS_METHOD(Integer_greaterthan) {
    INTEGER_BOOLEAN_OPERATION(>);
}

KITE_CLASS_METHOD(Integer_notequals) {
    INTEGER_BOOLEAN_OPERATION(!=);
}

KITE_CLASS_METHOD(Integer_not) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, !this->builtin_data.intvalue));
}

KITE_CLASS_METHOD(Integer_unaryplus) {
    KITE_NO_ARGS;
    int result = this->builtin_data.intvalue;
    if (result < 0) result = -result;
    kite_vm_return(thd, kite_new_integer(thd, result));
}

KITE_CLASS_METHOD(Integer_unaryminus) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_new_integer(thd, -this->builtin_data.intvalue));
}

KITE_CLASS_METHOD(Integer_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_string_object(thd, this));
}

KITE_CLASS_METHOD(Integer_int) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_reference_object(this));
}

KITE_CLASS_METHOD(Integer_float) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_float_object(thd, this));
}

KITE_CLASS_METHOD(Integer_bool) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_boolean_object(thd, this));
}

KITE_MODULE_INITIALIZER(integer) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "integer");
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ADD, 
        kite_new_method(thread, FUNC_COMPILED, Integer_add, 1));
    kite_add_operator(thread, newclass, OP_SUBTRACT, 
        kite_new_method(thread, FUNC_COMPILED, Integer_subtract, 1));
    kite_add_operator(thread, newclass, OP_MULTIPLY, 
        kite_new_method(thread, FUNC_COMPILED, Integer_multiply, 1));
    kite_add_operator(thread, newclass, OP_DIVIDE, 
        kite_new_method(thread, FUNC_COMPILED, Integer_divide, 1));
    kite_add_operator(thread, newclass, OP_MODULUS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_modulus, 1));
    kite_add_operator(thread, newclass, OP_LEFT_SHIFT, 
        kite_new_method(thread, FUNC_COMPILED, Integer_leftshift, 1));
    kite_add_operator(thread, newclass, OP_RIGHT_SHIFT, 
        kite_new_method(thread, FUNC_COMPILED, Integer_rightshift, 1));
    kite_add_operator(thread, newclass, OP_AND, 
        kite_new_method(thread, FUNC_COMPILED, Integer_and, 1));
    kite_add_operator(thread, newclass, OP_OR, 
        kite_new_method(thread, FUNC_COMPILED, Integer_or, 1));
    kite_add_operator(thread, newclass, OP_XOR, 
        kite_new_method(thread, FUNC_COMPILED, Integer_xor, 1));
    kite_add_operator(thread, newclass, OP_LESS_THAN, 
        kite_new_method(thread, FUNC_COMPILED, Integer_lessthan, 1));
    kite_add_operator(thread, newclass, OP_LESS_OR_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_lessequals, 1));
    kite_add_operator(thread, newclass, OP_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_equals, 1));
    kite_add_operator(thread, newclass, OP_GREATER_OR_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_greaterequals, 1));
    kite_add_operator(thread, newclass, OP_GREATER_THAN, 
        kite_new_method(thread, FUNC_COMPILED, Integer_greaterthan, 1));
    kite_add_operator(thread, newclass, OP_NOT_EQUALS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_notequals, 1));
    kite_add_operator(thread, newclass, OP_NOT, 
        kite_new_method(thread, FUNC_COMPILED, Integer_not, 1));
    kite_add_operator(thread, newclass, OP_UNARY_PLUS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_unaryplus, 1));
    kite_add_operator(thread, newclass, OP_UNARY_MINUS, 
        kite_new_method(thread, FUNC_COMPILED, Integer_unaryminus, 1));
    
    /* Add methods. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method(thread, FUNC_COMPILED, Integer_str, 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method(thread, FUNC_COMPILED, Integer_int, 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method(thread, FUNC_COMPILED, Integer_float, 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method(thread, FUNC_COMPILED, Integer_bool, 0));
        
    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "integer", 1);
    kite_set_property(thread, parent, "integer", newclass);
}

