/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netdb.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define CHECK_SOCKET_OPEN \
    if (this->builtin_data.intvalue < 0) { \
        kite_object_t *exc = \
            kite_new_exception(thd, "System.exceptions.SocketError", \
                               "Socket is not open"); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define THROW_SOCKET_ERROR \
    { \
        char buf[1024]; \
        strerror_r(errno, buf, 1024); \
        kite_object_t *exc =  \
            kite_new_exception(thd, "System.exceptions.SocketError",  \
                               buf); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define SET_SOCK_CONSTANT(c) \
    kite_add_property(thread, newclass, # c, 1); \
    kite_set_property(thread, newclass, # c, kite_new_integer(thread, c));  

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_construct) {
    kite_object_t *domain, *type, *protocol;

    KITE_GET_METHOD_ARGUMENT(domain, 1);
    KITE_GET_METHOD_ARGUMENT(type, 2);
    KITE_GET_METHOD_ARGUMENT(protocol, 3);

    if (!KITE_IS_TYPE(domain, OBJ_INTEGER) || 
        !KITE_IS_TYPE(type, OBJ_INTEGER) || 
        !KITE_IS_TYPE(protocol, OBJ_INTEGER))
    {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "All arguments must be integers");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    /*
     * Yes, we know it's not an integer, but we store the socket number
     * in intvalue.
     */
    KITE_SET_INTEGER(this, socket(KITE_GET_INTEGER(domain), 
                     KITE_GET_INTEGER(type), KITE_GET_INTEGER(protocol)));
    if (KITE_GET_INTEGER(this) < 0) 
        THROW_SOCKET_ERROR;
}

/*****************************************************************************
 * Write bytes to socket
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_write) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_string_object(thd, arg1);

    CHECK_SOCKET_OPEN;
    int ret = write(KITE_GET_INTEGER(this), KITE_GET_STRING_VALUE(bytes), 
                    KITE_GET_STRING_LENGTH(bytes));
    if (ret < 0)
        THROW_SOCKET_ERROR;

    kite_dereference_object(bytes);
    kite_vm_return(thd, kite_new_integer(thd, ret));
}

/*****************************************************************************
 * Read up to the given number of bytes
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_read) {
    kite_object_t *arg1 = args->builtin_data.listvalue.car;
    kite_object_t *bytes = kite_int_object(thd, arg1);
    char *buf = calloc(bytes->builtin_data.intvalue + 1, 1);
    int ret;

    CHECK_SOCKET_OPEN;

    ret = read(KITE_GET_INTEGER(this), buf, KITE_GET_STRING_LENGTH(bytes));
    if (ret < 0) {
        free(buf);
        THROW_SOCKET_ERROR;
    }

    kite_object_t *rval = kite_new_string(thd, buf);
    free(buf);
    kite_dereference_object(bytes);
    kite_vm_return(thd, rval);
}

/*****************************************************************************
 * Read a line of text from the socket.
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_readline) {
    char *buf = NULL;
    int ret, idx = 0;

    CHECK_SOCKET_OPEN;

    buf = calloc(1, 1);
    do {
        ret = read(KITE_GET_INTEGER(this), &buf[idx], 1);
        if (ret < 0) {
            free(buf);
            THROW_SOCKET_ERROR;
        } else if (ret == 0) {
            break;
        }
        idx++;
        buf = realloc(buf, idx + 1);
    } while(buf[idx - 1] != '\n');
    buf[idx] = 0;

    kite_object_t *rval = kite_new_string(thd, buf);
    free(buf);
    kite_vm_return(thd, rval);
}

/*****************************************************************************
 * Close given file
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_close) {
    KITE_NO_ARGS;

    CHECK_SOCKET_OPEN;
    int ret = close(KITE_GET_INTEGER(this));
    if (ret < 0)
        THROW_SOCKET_ERROR;

    KITE_SET_INTEGER(this, 0);
}

/*****************************************************************************
 * Accept new connection
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_accept) {
    struct sockaddr_storage addr;
    socklen_t len = sizeof(addr);
    KITE_NO_ARGS;

    CHECK_SOCKET_OPEN;
    int ret = accept(KITE_GET_INTEGER(this), 
        (struct sockaddr*)&addr, &len);
    if (ret < 0) 
        THROW_SOCKET_ERROR;

    char hostname[1024];
    kite_object_t *isUnix = kite_get_property(this, "unix_socket");
    
    if (!KITE_GET_INTEGER(isUnix))
        getnameinfo((struct sockaddr*)&addr, sizeof(addr), hostname, 1024, NULL,
                    0, NI_NUMERICHOST);
    else
        strcpy(hostname, "");

    kite_object_t *retval = kite_new_list(thd), *tmpret;
    kite_object_t *sock = kite_new_instance(thd, this->object_data.inherit_from);
    KITE_SET_INTEGER(sock, ret);
    retval->builtin_data.listvalue.car = sock;
    retval->builtin_data.listvalue.cdr = kite_new_list(thd);
    tmpret = retval->builtin_data.listvalue.cdr;
    tmpret->builtin_data.listvalue.car = kite_new_string(thd, hostname);
    tmpret->builtin_data.listvalue.cdr = kite_new_list(thd);
    tmpret = tmpret->builtin_data.listvalue.cdr;
    if (!KITE_GET_INTEGER(isUnix))
    {
        int port;
        
        switch(addr.ss_family)
        {
            case AF_INET:
            {
                struct sockaddr_in *sin = (struct sockaddr_in*)&addr;
                port = sin->sin_port;
                break;
            }
            case AF_INET6:
            {
                struct sockaddr_in6 *sin = (struct sockaddr_in6*)&addr;
                port = sin->sin6_port;
                break;
            }
        }
        tmpret->builtin_data.listvalue.car = kite_new_integer(thd, port);
    }
    else
        tmpret->builtin_data.listvalue.car = kite_new_integer(thd, 0);
        
    kite_vm_return(thd, retval);
}

/*****************************************************************************
 * Listen for connections on this socket
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_listen) {
    kite_object_t *backlog;

    KITE_GET_METHOD_ARGUMENT(backlog, 1);

    if (!KITE_IS_TYPE(backlog, OBJ_INTEGER))
    {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "Argument must be an integer");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    int ret = listen(KITE_GET_INTEGER(this), KITE_GET_INTEGER(backlog));
    if (ret < 0)
        THROW_SOCKET_ERROR;
}

/*****************************************************************************
 * Perform DNS lookup
 ****************************************************************************/
static int 
Socket_dnsLookup(kite_thread_t *thd, kite_object_t *this, int sock, 
                 char *host, int port, int mode)
{
    struct sockaddr *sa;
    struct hostent *hp;
    struct addrinfo hints;
    struct addrinfo *res;
    int ret = 0, error;
    size_t salen;
    char buf[128];
    int socktype;
    socklen_t len = sizeof(socklen_t);
    
    /* Unix domain sockets */
    kite_add_property(thd, this, "unix_socket", 0);
    if (host[0] == '/')
    {
        kite_set_property(thd, this, "unix_socket", kite_new_boolean(thd, TRUE));
                
        struct sockaddr_un local;
        local.sun_family = AF_UNIX;
        strcpy(local.sun_path, host);
        
        if (mode)
        {
            unlink(local.sun_path);
            ret = bind(sock, (struct sockaddr*)&local, sizeof(local));
        }
        else
        {
            ret = connect(sock, (struct sockaddr*)&local, sizeof(local));
        }
    }
    else
    {
        kite_set_property(thd, this, "unix_socket", kite_new_boolean(thd, FALSE));
        
        /* AF_INET[6] */
        memset(&hints, 0, sizeof(hints));
        sprintf(buf, "%d", port);
    
        getsockopt(sock, SOL_SOCKET, SO_TYPE, &socktype, &len);
    
        /* set-up hints structure */
        hints.ai_family = AF_UNSPEC;
        hints.ai_socktype = socktype;
        if (mode) hints.ai_flags = AI_PASSIVE;
        error = getaddrinfo(host, buf, &hints, &res);
        if (error)
        {
            kite_object_t *exc = 
                kite_new_exception(thd, "System.exceptions.DnsError", 
                                   gai_strerror(error));
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
            if (res) freeaddrinfo(res);
            return;
        }
        else 
        {
            struct addrinfo *top = res;
            while (res) {
                sa = res->ai_addr;
                salen = res->ai_addrlen;
            
                if (mode)
                {
                    /* bind() */
                    ret = bind(sock, sa, salen);
                    if (ret < 0) goto end_socket;
                }
                else
                {
                    /* connect() */
                    ret = connect(sock, sa, salen);
                    if (ret == 0) goto end_socket;
                }
            
                res = res->ai_next;
            }
end_socket:
            freeaddrinfo(top);
        }
    }
    return ret;
}

/*****************************************************************************
 * Bind address to socket
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_bind) {
    kite_object_t *host, *port;

    KITE_GET_METHOD_ARGUMENT(host, 1);
    KITE_GET_METHOD_ARGUMENT(port, 2);

    if (!KITE_IS_TYPE(host, OBJ_STRING) || !KITE_IS_TYPE(port, OBJ_INTEGER)) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "Arguments must be of the correct type (string, integer)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    int ret = Socket_dnsLookup(
        thd, this, KITE_GET_INTEGER(this), KITE_GET_STRING_VALUE(host), 
        KITE_GET_INTEGER(port), 1
    );

    if (ret < 0)
        THROW_SOCKET_ERROR;
}

/*****************************************************************************
 * Connect to server
 ****************************************************************************/
KITE_CLASS_METHOD(Socket_connect) {
    kite_object_t *host, *port;

    KITE_GET_METHOD_ARGUMENT(host, 1);
    KITE_GET_METHOD_ARGUMENT(port, 2);

    if (!KITE_IS_TYPE(host, OBJ_STRING) || !KITE_IS_TYPE(port, OBJ_INTEGER)) {
        kite_object_t *exc = 
            kite_new_exception(thd, "System.exceptions.TypeMismatch", 
                               "Arguments must be of the correct type (string, integer)");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }

    int ret = Socket_dnsLookup(
        thd, this, KITE_GET_INTEGER(this), KITE_GET_STRING_VALUE(host), 
        KITE_GET_INTEGER(port), 0
    );

    if (ret < 0)
        THROW_SOCKET_ERROR;
}

KITE_MODULE_INITIALIZER(socket) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits =
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = kite_new_class(thread, inherits, "socket");
    
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method(thread, FUNC_COMPILED, Socket_construct, 3));
    kite_add_method(thread, newclass, "accept", 
        kite_new_method(thread, FUNC_COMPILED, Socket_accept, 0));
    kite_add_method(thread, newclass, "bind", 
        kite_new_method(thread, FUNC_COMPILED, Socket_bind, 2));
    kite_add_method(thread, newclass, "connect", 
        kite_new_method(thread, FUNC_COMPILED, Socket_connect, 2));
    kite_add_method(thread, newclass, "listen", 
        kite_new_method(thread, FUNC_COMPILED, Socket_listen, 1));
    kite_add_method(thread, newclass, "read", 
        kite_new_method(thread, FUNC_COMPILED, Socket_read, 1));
    kite_add_method(thread, newclass, "readline", 
        kite_new_method(thread, FUNC_COMPILED, Socket_readline, 0));
    kite_add_method(thread, newclass, "write", 
        kite_new_method(thread, FUNC_COMPILED, Socket_write, 1));
    kite_add_method(thread, newclass, "close", 
        kite_new_method(thread, FUNC_COMPILED, Socket_close, 0));

    /* Add standard socket constants */
    SET_SOCK_CONSTANT(SOCK_DGRAM);
    SET_SOCK_CONSTANT(SOCK_STREAM);
    SET_SOCK_CONSTANT(SOCK_SEQPACKET);
    SET_SOCK_CONSTANT(AF_INET);
    SET_SOCK_CONSTANT(AF_INET6);
    SET_SOCK_CONSTANT(AF_UNSPEC);
    SET_SOCK_CONSTANT(AF_UNIX);

    /* Set destructor. */
    newclass->object_data.destructor =
        kite_new_method(thread, FUNC_COMPILED, Socket_close, 0);

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "socket", 1);
    kite_set_property(thread, parent, "socket", newclass);
}
