/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include "kite_object.h"
#include "vm/kite_vm.h"

#define TYPECHECK_RIGHT_OPERAND(t) \
    kite_object_t *rhs = args; \
    kite_object_t *exc; \
    if (rhs->type != t) { \
        exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                            "Second operand is not of the requested type."); \
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
        return; \
    }

#define TYPECHECK_OPERAND(v, t) \
    { \
        kite_object_t *rhs = v; \
        kite_object_t *exc; \
        if (rhs->type != t) { \
            exc = kite_new_exception(thd, "System.exceptions.TypeMismatch", \
                                "Second operand is not of the requested type."); \
            kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
            return; \
        } \
    }
    
#define TYPECHECK_METHOD_OPERAND TYPECHECK_RIGHT_OPERAND(OBJ_METHOD)

/*****************************************************************************
 * String representation of object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_str) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_object_name(thd, this));
}

/*****************************************************************************
 * Int representation of object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_int) {
    KITE_NO_ARGS;
    (void)this;
    kite_vm_return(thd, kite_new_integer(thd, 0));
}

/*****************************************************************************
 * Float representation of object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_float) {
    KITE_NO_ARGS;
    (void)this;
    kite_vm_return(thd, kite_new_float(thd, 0.0));
}

/*****************************************************************************
 * List representation of object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_list) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_list_object(thd, this));
}

/*****************************************************************************
 * Bool representation of object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_bool) {
    KITE_NO_ARGS;
    (void)this;
    kite_object_t *ret = kite_new_boolean(thd, FALSE);
    kite_vm_return(thd, ret);
}

/*****************************************************************************
 * Type of object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_type) {
    KITE_NO_ARGS;
    kite_vm_return(thd, kite_object_name(thd, this));
}

/*****************************************************************************
 * Lock mutex on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_lock) {
    KITE_NO_ARGS;
    pthread_mutex_lock(&this->object_data.mtx);
    kite_vm_return(thd, kite_new_null(thd));
}

/*****************************************************************************
 * Try to lock mutex on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_trylock) {
    KITE_NO_ARGS;
    int ret = !pthread_mutex_trylock(&this->object_data.mtx);
    kite_vm_return(thd, kite_new_boolean(thd, ret));
}

/*****************************************************************************
 * Unlock mutex on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_unlock) {
    KITE_NO_ARGS;
    pthread_mutex_unlock(&this->object_data.mtx);
    kite_vm_return(thd, kite_new_null(thd));
}

/*****************************************************************************
 * Share object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_share) {
    KITE_NO_ARGS;
    this->shareable = TRUE;
}

/*****************************************************************************
 * Unshare object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_unshare) {
    KITE_NO_ARGS;
    this->shareable = FALSE;
}

/*****************************************************************************
 * Common code to throw exception
 ****************************************************************************/
static void 
Object_throwNoIteratorSupported(kite_thread_t *thd, kite_object_t *this) {
    kite_object_t *name = kite_object_name(thd, this);
    char *msg = name->builtin_data.stringvalue.string;
    kite_object_t *exc;
    
    exc = kite_new_exception(thd, "System.exceptions.NoIteratorSupported", 
                             msg);
    kite_dereference_object(name);
                         
    kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
}

/*****************************************************************************
 * Reset iterator on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_reset) {
    KITE_NO_ARGS;
    Object_throwNoIteratorSupported(thd, this);
}

/*****************************************************************************
 * Current iterator location on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_cur) {
    KITE_NO_ARGS;
    Object_throwNoIteratorSupported(thd, this);
}

/*****************************************************************************
 * Next position in iterator on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_next) {
    KITE_NO_ARGS;
    Object_throwNoIteratorSupported(thd, this);
}

/*****************************************************************************
 * Next position in iterator on object (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_assignment) {
    kite_object_t *rhs = args;
    int i;
    
    assert(rhs && this->owner_thread->vm == rhs->owner_thread->vm);
    
    kite_dereference_object(this->parent);
    if (this->type == OBJ_LIST) {
        if (this->builtin_data.listvalue.car)
            kite_dereference_object(this->builtin_data.listvalue.car);
        if (this->builtin_data.listvalue.cdr)
            kite_dereference_object(this->builtin_data.listvalue.cdr);
    } else if (this->type == OBJ_STRING) {
        free(this->builtin_data.stringvalue.string);
    }
    
    this->type = rhs->type;
    this->parent = kite_reference_object(rhs->parent);
    memcpy(&this->builtin_data, &rhs->builtin_data, sizeof(this->builtin_data));
    if (this->type == OBJ_LIST) {
        this->builtin_data.listvalue.car = 
            kite_reference_object(this->builtin_data.listvalue.car);
        this->builtin_data.listvalue.cdr = 
            kite_reference_object(this->builtin_data.listvalue.cdr);
        this->builtin_data.listvalue.iterator = NULL;
    } else if (this->type == OBJ_STRING) {
        this->builtin_data.stringvalue.string = 
            strdup(this->builtin_data.stringvalue.string);
    } else if ((this->type == OBJ_METHOD || this->type == OBJ_CLASS) && 
               this->builtin_data.funcvalue.functype == FUNC_BYTECODE &&
               this->builtin_data.funcvalue.funcptr != NULL) {
        this->builtin_data.funcvalue.funcptr = 
            kite_copy_instruction_list(thd, this->builtin_data.funcvalue.funcptr);
    }
    
    free(this->object_data.name);
    if (rhs->object_data.name) {
        this->object_data.name = strdup(rhs->object_data.name);
    } else {
        this->object_data.name = NULL;
    }
    if (this->object_data.inherit_from) {
        kite_dereference_object(this->object_data.inherit_from);
    }
    this->object_data.inherit_from = 
        kite_reference_object(rhs->object_data.inherit_from);
    for(i = 0; i < NUM_OPERATORS; i++) {
        kite_dereference_object(this->object_data.operators[i]);
        this->object_data.operators[i] = 
            kite_reference_object(rhs->object_data.operators[i]);
    }
    if (this->object_data.properties) {
        kite_destruct_symtab(thd, &this->object_data.properties, FALSE);
        this->object_data.properties = NULL;
    }
    if (rhs->object_data.properties)
        kite_copy_symtab(&this->object_data.properties, 
                        rhs->object_data.properties);
    if (this->object_data.destructor) {
        kite_dereference_object(this->object_data.destructor);
    }
    this->object_data.destructor = 
        kite_reference_object(rhs->object_data.destructor);
    this->object_data.creatable = rhs->object_data.creatable;
    
    kite_vm_return(thd, this);
}

/*****************************************************************************
 * Print object on standard output (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_print) {
    KITE_NO_ARGS;
    kite_object_t *str = kite_string_object(thd, this);
    printf("%s\n", str->builtin_data.stringvalue.string);
    kite_dereference_object(str);
}

/*****************************************************************************
 * Return object's docstring.
 ****************************************************************************/
KITE_CLASS_METHOD(Object_docString) {
    KITE_NO_ARGS;
    kite_object_t *str = kite_new_string(thd, kite_get_docstring(this));
    kite_vm_push(thd, str);
    kite_dereference_object(str);
}

static void 
expandProperties(kite_thread_t *thd, kite_symtab_t *symt, kite_object_t **list)
{
    if (symt->left) {
        expandProperties(thd, symt->left, list);
        while(*list && (*list)->builtin_data.listvalue.cdr) 
            list = &(*list)->builtin_data.listvalue.cdr;
    }
    
    if (!symt->value || (symt->value && symt->value->type != OBJ_METHOD)) {
try_again2:
        if (!*list) {
            *list = kite_new_list(thd);
        } else {
            list = &(*list)->builtin_data.listvalue.cdr;
            goto try_again2;
        }
        (*list)->builtin_data.listvalue.car = kite_new_string(thd, symt->name);
        list = &(*list)->builtin_data.listvalue.cdr;
    }
    
    if (symt->right) {
        expandProperties(thd, symt->right, list);
    }
}

/*****************************************************************************
 * Return object's list of valid properties.
 ****************************************************************************/
KITE_CLASS_METHOD(Object_listProperties) {
    KITE_NO_ARGS;
    kite_object_t *ret = NULL;
    
    if (this->object_data.properties) {
        expandProperties(thd, this->object_data.properties, &ret);
        if (!ret)
            ret = kite_new_list(thd);
    } else {
        ret = kite_new_list(thd);
    }
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

static void 
expandOverloadedMethods(kite_thread_t *thd, kite_symtab_t *symt, char *name, kite_object_t **list)
{
    int r = strcmp(name, symt->name);
    if (r != 0) return;
    
    if (symt->left) {
        expandOverloadedMethods(thd, symt->left, name, list);
        while(*list && (*list)->builtin_data.listvalue.cdr) 
            list = &(*list)->builtin_data.listvalue.cdr;
    }
    
    if (!r) {
        if (!*list) {
            *list = kite_new_list(thd);
        } else {
            list = &(*list)->builtin_data.listvalue.cdr;
        }
        (*list)->builtin_data.listvalue.car = kite_new_integer(thd, symt->global);
        list = &(*list)->builtin_data.listvalue.cdr;
    }
    
    if (symt->right) {
        expandOverloadedMethods(thd, symt->right, name, list);
    }
}

static void 
expandMethods(kite_thread_t *thd, kite_symtab_t *symt, kite_object_t **list)
{
    if (symt->left) {
        kite_symtab_t *lsymt = symt->left;
        while(lsymt && !strcmp(lsymt->name, symt->name))
            lsymt = lsymt->left;
            
        if (lsymt) expandMethods(thd, lsymt, list);
        while(*list && (*list)->builtin_data.listvalue.cdr) 
            list = &(*list)->builtin_data.listvalue.cdr;
    }
    
    if (symt->value && symt->value->type == OBJ_METHOD) {
try_again:
        if (!*list) {
            *list = kite_new_list(thd);
        } else {
            list = &(*list)->builtin_data.listvalue.cdr;
            goto try_again;
        }
        (*list)->builtin_data.listvalue.car = kite_new_string(thd, symt->name);
        (*list)->builtin_data.listvalue.cdr = kite_new_list(thd);
        kite_object_t *nxt = (*list)->builtin_data.listvalue.cdr;
        expandOverloadedMethods(thd, symt, symt->name, &nxt->builtin_data.listvalue.car);
        list = &nxt->builtin_data.listvalue.cdr;
    }
    
    if (symt->right) {
        kite_symtab_t *rsymt = symt->right;
        while(rsymt && !strcmp(rsymt->name, symt->name))
            rsymt = rsymt->right;
            
        if (rsymt) expandMethods(thd, rsymt, list);
    }
}

/*****************************************************************************
 * Return object's list of valid methods.
 ****************************************************************************/
KITE_CLASS_METHOD(Object_listMethods) {
    KITE_NO_ARGS;
    kite_object_t *ret = NULL;
    
    if (this->object_data.properties) {
        expandMethods(thd, this->object_data.properties, &ret);
        if (!ret)
            ret = kite_new_list(thd);
    } else {
        ret = kite_new_list(thd);
    }
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Get property's docstring
 ****************************************************************************/
KITE_CLASS_METHOD(Object_getMethod) {
    kite_symtab_t *symt;
    kite_object_t *ret, *name, *number;
    
    KITE_GET_METHOD_ARGUMENT(name, 1);
    KITE_GET_METHOD_ARGUMENT(number, 2);
    TYPECHECK_OPERAND(name, OBJ_STRING);
    TYPECHECK_OPERAND(number, OBJ_INTEGER);
    
    KITE_FIND_METHOD(symt, this, KITE_GET_STRING_VALUE(name), KITE_GET_INTEGER(number));
    if (symt) {
        kite_vm_push(thd, symt->value);
    } else {
        ret = kite_new_null(thd);
        kite_vm_push(thd, ret);
        kite_dereference_object(ret);
    }
}

/*****************************************************************************
 * Get destructor's method object
 ****************************************************************************/
KITE_CLASS_METHOD(Object_getDestructor) {
    if (this->object_data.destructor) {
        kite_vm_push(thd, this->object_data.destructor);
    } else {
        kite_object_t *ret = kite_new_null(thd);
        kite_vm_push(thd, ret);
        kite_dereference_object(ret);
    }
}

/*****************************************************************************
 * Get whether this object is a class
 ****************************************************************************/
KITE_CLASS_METHOD(Object_isClass) {
    kite_object_t *ret = kite_new_boolean(thd, this->type == OBJ_CLASS);
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Get object this inherits.
 ****************************************************************************/
KITE_CLASS_METHOD(Object_getBaseObject) {
    if (this->object_data.inherit_from) {
        kite_vm_push(thd, this->object_data.inherit_from);
    } else {
        kite_object_t *ret = kite_new_null(thd);
        kite_vm_push(thd, ret);
        kite_dereference_object(ret);
    }
}

/*****************************************************************************
 * Get property's docstring
 ****************************************************************************/
KITE_CLASS_METHOD(Object_getPropertyString) {
    kite_symtab_t *symt;
    kite_object_t *ret;
    
    KITE_GET_METHOD_ARGUMENT(ret, 1);
    TYPECHECK_OPERAND(ret, OBJ_STRING);
    KITE_FIND_ANY_IN_SYMTAB(symt, this->object_data.properties, KITE_GET_STRING_VALUE(ret));
    
    if (symt) {
        if (symt->docstring) ret = kite_new_string(thd, symt->docstring);
        else if (symt->value && symt->value->object_data.docstring) {
            ret = kite_new_string(thd, symt->value->object_data.docstring);
        } else {
            ret = kite_new_string(thd, "");
        }
    } else {
        ret = kite_new_string(thd, "");
    }
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Get property's value
 ****************************************************************************/
KITE_CLASS_METHOD(Object_getProperty) {
    kite_symtab_t *symt;
    kite_object_t *ret;
    
    KITE_GET_METHOD_ARGUMENT(ret, 1);
    TYPECHECK_OPERAND(ret, OBJ_STRING);
    KITE_FIND_ANY_IN_SYMTAB(symt, this->object_data.properties, KITE_GET_STRING_VALUE(ret));
    
    if (symt) {
        if (symt->value)
            ret = kite_reference_object(symt->value);
        else
            ret = kite_new_null(thd);
    } else {
        ret = kite_new_null(thd);
    }
    
    kite_vm_push(thd, ret);
    kite_dereference_object(ret);
}

/*****************************************************************************
 * Map operator (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_map) {
    kite_object_t *ret, *cur, *curret, *methodargs;
    TYPECHECK_METHOD_OPERAND;
    
    kite_vm_call_method(thd, this, "reset", kite_new_list(thd), 1);
    kite_vm_call_method(thd, this, "cur", kite_new_list(thd), 1);
    cur = kite_vm_pop(thd);
    
    ret = curret = kite_new_list(thd);
    while(cur->type != OBJ_NULL) {
        methodargs = kite_list_object(thd, cur);
        kite_vm_call_object(thd, this, rhs, methodargs);
        curret->builtin_data.listvalue.car = kite_vm_pop(thd);    
        kite_dereference_object(methodargs);

        kite_dereference_object(cur);
        kite_vm_call_method(thd, this, "next", kite_new_list(thd), 1);
        cur = kite_vm_pop(thd);
        if (cur->builtin_data.intvalue) {
            curret->builtin_data.listvalue.cdr = kite_new_list(thd);
        }
        kite_dereference_object(cur);
        
        kite_vm_call_method(thd, this, "cur", kite_new_list(thd), 1);
        cur = kite_vm_pop(thd);
        curret = curret->builtin_data.listvalue.cdr;
    }
    
    kite_dereference_object(cur);
    kite_vm_return(thd, ret);
}

/*****************************************************************************
 * Reduce operator (default implementation)
 ****************************************************************************/
KITE_CLASS_METHOD(Object_reduce) {
    kite_object_t *ret, *cur, *methodargs;
    TYPECHECK_METHOD_OPERAND;
    
    kite_vm_call_method(thd, this, "reset", kite_new_list(thd), 1);
    kite_vm_call_method(thd, this, "cur", kite_new_list(thd), 1);
    cur = kite_vm_pop(thd);
    
    ret = kite_new_null(thd);
    while(cur->type != OBJ_NULL) {
        methodargs = kite_new_list(thd);
        methodargs->builtin_data.listvalue.car = kite_reference_object(ret);
        methodargs->builtin_data.listvalue.cdr = 
            kite_list_object(thd, cur);
        kite_vm_call_object(thd, this, rhs, methodargs);
        kite_dereference_object(ret);
        ret = kite_vm_pop(thd);
        
        kite_dereference_object(methodargs);
        kite_dereference_object(cur);
        
        kite_vm_call_method(thd, this, "next", kite_new_list(thd), 1);
        kite_vm_call_method(thd, this, "cur", kite_new_list(thd), 1);
        cur = kite_vm_pop(thd);
    }
    
    kite_dereference_object(cur);
    kite_vm_return(thd, ret);
}

/*****************************************************************************
 * Get operator method
 ****************************************************************************/
KITE_CLASS_METHOD(Object_getOperatorMethod)
{
    kite_object_t *obj;
    
    KITE_GET_METHOD_ARGUMENT(obj, 1);
    TYPECHECK_OPERAND(obj, OBJ_INTEGER);
    
    int opint = KITE_GET_INTEGER(obj);
    if (opint < 0 || opint >= NUM_OPERATORS) {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.InvalidArgument",
                            "Operator is not valid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    obj = this->object_data.operators[opint];
    if (obj != NULL) {
        kite_vm_push(thd, obj);
    } else {
        obj = kite_new_null(thd);
        kite_vm_push(thd, obj);
        kite_dereference_object(obj);
    }
}

/*****************************************************************************
 * Convert operator number to string
 ****************************************************************************/
KITE_CLASS_METHOD(Object_operators_toStr)
{
    kite_object_t *obj;
    
    KITE_GET_METHOD_ARGUMENT(obj, 1);
    TYPECHECK_OPERAND(obj, OBJ_INTEGER);
    
    int opint = KITE_GET_INTEGER(obj);
    if (opint < 0 || opint >= NUM_OPERATORS) {
        kite_object_t *exc;
        exc = kite_new_exception(thd, "System.exceptions.InvalidArgument",
                            "Operator is not valid.");
        kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE);
        return;
    }
    
    obj = kite_new_string(thd, OP_TO_STRING(opint));
    kite_vm_push(thd, obj);
    kite_dereference_object(obj);
}

extern char *OPERATOR_NAME_STRINGS[];

KITE_MODULE_INITIALIZER(object) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *newclass = kite_new_class(thread, NULL, "object");
    
    /* Add to parent package. */
    kite_add_property(thread, parent, "object", 1);
    kite_set_property(thread, parent, "object", newclass);
    
    /*
     * Add operators
     */
    kite_add_operator(thread, newclass, OP_ASSIGNMENT, 
        kite_new_method(thread, FUNC_COMPILED, Object_assignment, 1));
    kite_add_operator(thread, newclass, OP_MAP, 
        kite_new_method(thread, FUNC_COMPILED, Object_map, 1));
    kite_add_operator(thread, newclass, OP_REDUCE, 
        kite_new_method(thread, FUNC_COMPILED, Object_reduce, 1));
        
    /* Add methods common to all classes. */
    kite_add_method(thread, newclass, "str", 
        kite_new_method(thread, FUNC_COMPILED, Object_str, 0));
    kite_add_method(thread, newclass, "int", 
        kite_new_method(thread, FUNC_COMPILED, Object_int, 0));
    kite_add_method(thread, newclass, "float", 
        kite_new_method(thread, FUNC_COMPILED, Object_float, 0));
    kite_add_method(thread, newclass, "list", 
        kite_new_method(thread, FUNC_COMPILED, Object_list, 0));
    kite_add_method(thread, newclass, "bool", 
        kite_new_method(thread, FUNC_COMPILED, Object_bool, 0));
    kite_add_method(thread, newclass, "type", 
        kite_new_method(thread, FUNC_COMPILED, Object_type, 0));
        
    /* Object shareability properties */
    kite_add_method(thread, newclass, "lock", 
        kite_new_method(thread, FUNC_COMPILED, Object_lock, 0));
    kite_add_method(thread, newclass, "trylock", 
        kite_new_method(thread, FUNC_COMPILED, Object_trylock, 0));
    kite_add_method(thread, newclass, "unlock", 
        kite_new_method(thread, FUNC_COMPILED, Object_unlock, 0));
    kite_add_method(thread, newclass, "share", 
        kite_new_method(thread, FUNC_COMPILED, Object_share, 0));
    kite_add_method(thread, newclass, "unshare", 
        kite_new_method(thread, FUNC_COMPILED, Object_unshare, 0));

    kite_add_method(thread, newclass, "doc",
        kite_new_method(thread, FUNC_COMPILED, Object_docString, 0));
    kite_add_method(thread, newclass, "list_properties",
        kite_new_method(thread, FUNC_COMPILED, Object_listProperties, 0));
    kite_add_method(thread, newclass, "list_methods",
        kite_new_method(thread, FUNC_COMPILED, Object_listMethods, 0));
    kite_add_method(thread, newclass, "get_method",
        kite_new_method(thread, FUNC_COMPILED, Object_getMethod, 2));
    kite_add_method(thread, newclass, "get_destructor",
        kite_new_method(thread, FUNC_COMPILED, Object_getDestructor, 0));
    kite_add_method(thread, newclass, "is_class",
        kite_new_method(thread, FUNC_COMPILED, Object_isClass, 0));
    kite_add_method(thread, newclass, "get_property",
        kite_new_method(thread, FUNC_COMPILED, Object_getProperty, 1));
    kite_add_method(thread, newclass, "get_property_string",
        kite_new_method(thread, FUNC_COMPILED, Object_getPropertyString, 1));
    kite_add_method(thread, newclass, "get_base_object",
        kite_new_method(thread, FUNC_COMPILED, Object_getBaseObject, 0));
    kite_add_method(thread, newclass, "get_operator_method",
        kite_new_method(thread, FUNC_COMPILED, Object_getOperatorMethod, 1));

    kite_add_method(thread, newclass, "reset", 
        kite_new_method(thread, FUNC_COMPILED, Object_reset, 0));
    kite_add_method(thread, newclass, "cur", 
        kite_new_method(thread, FUNC_COMPILED, Object_cur, 0));
    kite_add_method(thread, newclass, "next", 
        kite_new_method(thread, FUNC_COMPILED, Object_next, 0));
    
    kite_add_method(thread, newclass, "print", 
        kite_new_method(thread, FUNC_COMPILED, Object_print, 0));
    
    /* Operator array */
    kite_object_t *oparray = kite_new_class(thread, newclass, "operators");
    kite_add_property(thread, newclass, "operators", 1);
    kite_set_property(thread, newclass, "operators", oparray);
    int i;
    for (i = 0; i <= NUM_OPERATORS; i++) {
        kite_add_property(thread, oparray, OPERATOR_NAME_STRINGS[i], 1);
        kite_set_property(thread, oparray, OPERATOR_NAME_STRINGS[i],
            kite_new_integer(thread, i));
    }
    kite_add_method(thread, oparray, "to_str",
        kite_new_method(thread, FUNC_COMPILED, Object_operators_toStr, 1));
}
