/*****************************************************************************
* Copyright (c) 2008, Mooneer Salem
* All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*     * Redistributions of source code must retain the above copyright
*       notice, this list of conditions and the following disclaimer.
*     * Redistributions in binary form must reproduce the above copyright
*       notice, this list of conditions and the following disclaimer in the
*       documentation and/or other materials provided with the distribution.
*     * Neither the name of the Kite Language organization nor the
*       names of its contributors may be used to endorse or promote products
*       derived from this software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
* EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
* DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
* DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "microregex_internal.h"

microregex_nfa_t microregex_parse_regex(microregex_t, char *, char **, int);
void microregex_destroy_nfa(microregex_nfa_t *);

static void
microregex_nfa_append(microregex_nfa_t *nfa, microregex_nfa_t end)
{
    if (!*nfa)
    {
        *nfa = end;
        return;
    }

    if ((*nfa)->state_type != PASS && (*nfa)->state_type != REGEX_END && (*nfa)->next1 != end)
        microregex_nfa_append(&(*nfa)->next1, end);
    if ((*nfa)->state_type == ALTERATION || (*nfa)->state_type == COUNTED_LOOP)
        if ((*nfa)->next2 != end)
            microregex_nfa_append(&(*nfa)->next2, end);
}

static microregex_nfa_t
microregex_nfa_alteration(microregex_nfa_t next1, microregex_nfa_t next2)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = ALTERATION;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = next1;
    ret->next2 = next2;
    return ret;
}

static microregex_nfa_t
microregex_nfa_assert_beginning()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = ASSERT_BEGINNING;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_assert_end()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = ASSERT_END;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_counted_begin()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = COUNTED_BEGIN;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_counted_loop(int from, int to)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = COUNTED_LOOP;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->loop_from = from;
    ret->loop_to = to;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_counted_end()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = COUNTED_END;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_push_submatch()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = PUSH_SUBMATCH;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_pop_submatch()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = POP_SUBMATCH;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_end()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = REGEX_END;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_pass()
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    ret->state_type = PASS;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;
    return ret;
}

static microregex_nfa_t
microregex_nfa_oneplus(microregex_nfa_t one, int greedy)
{
    microregex_nfa_t pass = microregex_nfa_pass();
    microregex_nfa_t ret = 
        microregex_nfa_alteration(
            greedy ? pass : NULL,
            greedy ? NULL : pass
        );
    microregex_nfa_append(&one, ret);
    pass->next1 = one;
    return one;
}

static microregex_nfa_t
microregex_nfa_zeroplus(microregex_nfa_t one, int greedy)
{
    microregex_nfa_t ret = 
        microregex_nfa_alteration(
            greedy ? one : NULL, 
            greedy ? NULL : one
        );
    microregex_nfa_t pass = microregex_nfa_pass();
    microregex_nfa_append(&one, pass);
    pass->next1 = ret;
    return ret;
}

static microregex_nfa_t
microregex_nfa_zeroone(microregex_nfa_t one, int greedy)
{
    microregex_nfa_t pass = microregex_nfa_pass();
    microregex_nfa_t ce = microregex_nfa_counted_end();
    microregex_nfa_t ret = 
        microregex_nfa_alteration(
            greedy ? one : ce, 
            greedy ? ce : one
        );
    microregex_nfa_append(&one, pass);
    pass->next1 = ce;
    return ret;
}

static char *
microregex_nfa_charliteral_process(char *regex, char *buf)
{
    int i, j;

    memset(buf, 0, 256);

    if (*regex == '.')
    {
        for (i = 1; i <= 255; i++)
        {
            buf[i - 1] = i;
        }
    }
    else if (*regex == '\\' && *(regex + 1))
    {
        regex++;
        switch(*regex)
        {
            case 't':
            {
                buf[0] = '\t';
                break;
            }
            case 'n':
            {
                buf[0] = '\n';
                break;
            }
            case 'r':
            {
                buf[0] = '\r';
                break;
            }
            case 'f':
            {
                buf[0] = '\f';
                break;
            }
            case 'a':
            {
                buf[0] = '\a';
                break;
            }
            case 'e':
            {
                buf[0] = '\e';
                break;
            }
            case '0':
            {
                buf[0] = (char)strtol(regex, &regex, 8);
                break;
            }
            case 'x':
            {
                buf[0] = (char)strtol(regex, &regex, 16);
                break;
            }
            case 'w':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (isalnum(i) || ((char)i) == '_')
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'W':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (!isalnum(i) && (!((char)i) == '_'))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 's':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (isspace(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'S':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (!isspace(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'd':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (isdigit(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            case 'D':
            {
                for (i = 1, j = 0; i <= 255; i++)
                {
                    if (!isdigit(i))
                    {
                        buf[j++] = (char)i;
                    }
                }
                break;
            }
            default:
            {
                buf[0] = *regex;
                break;
            }
        }
    }
    else
    {
        buf[0] = *regex;
    }

    return ++regex;
}

static microregex_nfa_t
microregex_nfa_charliteral(char *regex, char **end_location)
{
    microregex_nfa_t ret = (microregex_nfa_t)malloc(sizeof(struct microregex_nfa_t));
    char tmp_buf[256];

    ret->state_type = CHARACTER_CLASS;
    ret->character_class = NULL;
    ret->invert = 0;
    ret->next1 = NULL;
    ret->next2 = NULL;

    memset(tmp_buf, 0, 256);
    ret->character_class = malloc(1);
    ret->character_class[0] = 0;

    if (*regex == '[')
    {
        char *tmp = regex + 1;
        if (*tmp == '^') 
        {
            ret->invert++;
            tmp++;
        }
        regex = tmp;
        while(*tmp && *tmp != ']')
        {
            if (strlen(tmp_buf) == 1 && *tmp == '-' && *(tmp + 1) != ']')
            {
                char from = tmp_buf[0];
                char b[2];
                tmp = microregex_nfa_charliteral_process(tmp + 1, tmp_buf);
                if (strlen(tmp_buf) == 1 && tmp_buf[0] > from)
                {
                    for (; from <= tmp_buf[0]; from++)
                    {
                        ret->character_class = 
                            realloc(
                                ret->character_class, 
                                strlen(ret->character_class) + 2
                            );
                        b[0] = from;
                        b[1] = 0;
                        strcat(ret->character_class, b);
                    }
                }
                else
                {
                    ret->character_class = 
                        realloc(
                            ret->character_class, 
                            strlen(ret->character_class) + 3 + strlen(tmp_buf)
                        );
                    strcat(ret->character_class, tmp_buf);
                    b[0] = from;
                    b[1] = 0;
                    strcat(ret->character_class, b);
                    b[0] = '-';
                    strcat(ret->character_class, b);
                }
                continue;
            }
            else
            {
                ret->character_class = 
                    realloc(
                        ret->character_class, 
                        strlen(ret->character_class) + strlen(tmp_buf) + 1
                    );
                strcat(ret->character_class, tmp_buf);
            }

            tmp = microregex_nfa_charliteral_process(tmp, tmp_buf);
            if (!*tmp)
            {
                ret->character_class = 
                    realloc(
                        ret->character_class, 
                        strlen(ret->character_class) + strlen(tmp_buf) + 1
                    );
                strcat(ret->character_class, tmp_buf);
            }
        }

        regex = tmp;
    }
    else
    {
        regex = microregex_nfa_charliteral_process(regex, tmp_buf);
        ret->character_class = 
            realloc(
                ret->character_class, 
                strlen(ret->character_class) + strlen(tmp_buf) + 1
            );
        strcat(ret->character_class, tmp_buf);
        regex--;
    }

    if (end_location) *end_location = regex;
    return ret;
}

/**
 * Create regular expression object.
 * @param regex The regular expression to parse.
 * @return The created regular expression object.
 */
microregex_t
microregex_create(char *regex)
{
    microregex_t ret = (microregex_t)malloc(sizeof(struct microregex_t));
    ret->paren_count = 0;
    ret->regex = regex;
    ret->err_string = NULL;
    ret->nfa = microregex_parse_regex(ret, regex, NULL, 0);
    microregex_nfa_append(&ret->nfa, microregex_nfa_end());

    return ret;
}

/**
 * Destroy regular expression object.
 * @param regex_obj The regular expression object to destroy.
 */
void
microregex_destroy_nfa(microregex_nfa_t *regex_obj)
{
    if (regex_obj == NULL || (*regex_obj) == NULL) return;

    if ((*regex_obj)->state_type != PASS)
    {
        microregex_destroy_nfa(&(*regex_obj)->next1);
        microregex_destroy_nfa(&(*regex_obj)->next2);
    }

    free((*regex_obj)->character_class);
    free(*regex_obj);
    *regex_obj = NULL;
}

/**
 * Destroy regular expression object.
 * @param regex_obj The regular expression object to destroy.
 */
void
microregex_destroy(microregex_t *regex_obj)
{
    if (regex_obj == NULL || (*regex_obj) == NULL) return;

    if ((*regex_obj)->err_string) free((*regex_obj)->err_string);
    if ((*regex_obj)->nfa) microregex_destroy_nfa(&(*regex_obj)->nfa);
    free(*regex_obj);
    *regex_obj = NULL;
}

/**
 * Set case insensitivity of regular expression.
 * @param regex_obj The regular expression object to edit.
 * @param val Value to set flag to.
 */
void microregex_set_case_insensitivity(microregex_t regex_obj, int val)
{
    regex_obj->case_insensitive = val;
}

/**
 * Parse regular expression into an NFA.
 * @param regex_obj The regular expression object with state.
 * @param regex The regular expression to parse.
 * @param end_char The location where it stopped parsing
 * @param paren_start The starting nesting level.
 * @return The created NFA object.
 */
microregex_nfa_t
microregex_parse_regex(microregex_t regex_obj, char *regex, char **end_char, int paren_start)
{
    int cur_paren = regex_obj->paren_count;
    microregex_nfa_t nfa = NULL;
    microregex_nfa_t prev = NULL;
    char *end_location;
    int greedy;

    while(*regex && regex_obj->paren_count >= cur_paren) 
    {
        if (prev && *regex != '+' && *regex != '?' && *regex != '*' && *regex != '{')
        {
            microregex_nfa_append(&nfa, prev);
            prev = NULL;
        }

        switch(*regex)
        {
            case '(':
            {
                regex_obj->paren_count++;
                microregex_nfa_append(
                    &prev,
                    microregex_nfa_push_submatch()
                );
                microregex_nfa_append(
                    &prev,
                    microregex_parse_regex(
                        regex_obj,
                        regex + 1,
                        &end_location,
                        paren_start + 1
                    )
                );

                regex = end_location;
                break;
            }
            case ')':
            {
                if (regex_obj->paren_count == paren_start)
                {
                    microregex_nfa_append(
                        &prev,
                        microregex_nfa_pop_submatch()
                    );
                }
                
                regex_obj->paren_count--;
                regex--; /* We increment at the bottom of the loop */
                break;
            }
            case '|':
            {
                /* Alternation */
                microregex_nfa_t alt;

                regex_obj->paren_count++;
                
                alt = microregex_nfa_alteration(
                    nfa,
                    microregex_parse_regex(
                        regex_obj,
                        regex + 1,
                        &end_location,
                        paren_start
                    )
                );

                regex = end_location - 1;
                
                nfa = alt;
                break;
            }
            case '+':
            {
                greedy = (*(++regex) != '?');

                /* One or more */
                prev =
                    microregex_nfa_oneplus(
                        prev,
                        greedy
                    );

                if (*regex != '?') regex--;
                break;
            }
            case '*':
            {
                greedy = (*(++regex) != '?');

                /* Zero or more */
                prev = 
                    microregex_nfa_zeroplus(
                        prev,
                        greedy
                    );

                if (*regex != '?') regex--;
                break;
            }
            case '?':
            {
                greedy = (*(++regex) != '?');

                /* Zero or one */
                prev =    
                    microregex_nfa_zeroone(
                        prev,
                        greedy
                    );

                if (*regex != '?') regex--;
                break;
            }
            case '{':
            {
                /* Counted repetition */
                char *cur_pos = regex;
                int from = strtol(regex + 1, &regex, 10);
                int to;
                if (!*regex || (*regex != ',' && *regex != '}'))
                {
                    regex = cur_pos;
                    goto default_character;
                }
                else if (*regex == ',')
                {
                    to = strtol(regex + 1, &regex, 10);
                    if (*regex != '}')
                    {
                        regex = cur_pos;
                        goto default_character;
                    }
                    else if (to == 0)
                    {
                        to = INT_MAX;
                    }
                }
                else
                {
                    to = from;
                }

                microregex_nfa_t tmpnfa = microregex_nfa_counted_begin();
                microregex_nfa_t counted_loop = microregex_nfa_counted_loop(from, to);
                greedy = (*(++regex) != '?');
                microregex_nfa_t tmppass = microregex_nfa_pass();
                tmppass->next1 = counted_loop;

                counted_loop->next1 = greedy ? prev : microregex_nfa_counted_end();
                counted_loop->next2 = greedy ? microregex_nfa_counted_end() : prev;
                microregex_nfa_append(
                    &prev,
                    tmppass
                );

                tmpnfa->next1 = counted_loop;

                prev = tmpnfa;
                if (greedy) regex--;
                break;
            }
            case '^':
            {
                /* Beginning of string */
                microregex_nfa_append(
                    &nfa,
                    microregex_nfa_assert_beginning()
                );
                break;
            }
            case '$':
            {
                /* End of string */
                microregex_nfa_append(
                    &nfa,
                    microregex_nfa_assert_end()
                );
                break;
            }
            case '[':
            default:
            {
default_character:
                /* Character class/literal */
                prev = microregex_nfa_charliteral(
                    regex,
                    &end_location
                );

                regex = end_location;
            }
        };

        regex++;
    }

    if (prev) microregex_nfa_append(&nfa, prev);
    if (end_char) *end_char = regex;
    return nfa;
}
