#ifndef KITE_VM__KITEVM_H
#define KITE_VM__KITEVM_H
/*****************************************************************************
 * Copyright (c) 2007, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <setjmp.h>
#include <pthread.h>
#include <signal.h>
#include <assert.h>
#include "kite_opcodes.h"

struct kite_object_t;
struct kite_symtab_t;

/*
 * Stack representation
 */
struct kite_stack_entry_t;
typedef struct kite_stack_t
{
    size_t allocated;
    int length;
    struct kite_stack_entry_t *stack;
} kite_stack_t;

typedef struct kite_stack_entry_t
{
    int reference;
	void *obj;
} kite_stack_entry_t;
    
/*
 * Entry on function call stack.
 */
typedef struct kite_funccall_info_t
{
    char *file;
    int line;
    int sys_or_user;
    struct kite_symtab_t *locals;
    struct kite_object_t *this;
} kite_funccall_info_t;
    
/*
 * Kite thread object. This object represents a single thread in the Kite
 * threading system (and its current state)
 */
typedef struct kite_thread_t
{
    struct kite_thread_t **entry;
    struct kite_vm_t *vm;
    kite_stack_t *running_stack;
    kite_stack_t *func_stack;
    struct kite_object_t *exception;
    void *start;
    pthread_t thread;
} kite_thread_t;

/*
 * Kite list object. 
 */
typedef struct kite_list_t
{
    void *obj;
    struct kite_list_t *prev, *next;
} kite_list_t;

/*
 * Kite VM object. This object is how other developers will access
 * Kite internals.
 */
typedef struct kite_vm_t
{
	struct kite_object_t *root_package;
	struct kite_object_t *true_cached;
	struct kite_object_t *false_cached;
	struct kite_object_t *null_cached;
	struct kite_object_t *int_cached;
    struct kite_symtab_t *loader_refs;
    int numthreads;
    kite_thread_t **thread_list;
    kite_list_t *gc_begin, *gc_end; /* garbage collector queue */
    pthread_mutex_t gc_mtx;
    char **args;
} kite_vm_t;

/*
 * Kite compiler state object
 */
typedef struct kite_compiler_t
{
    kite_thread_t *thd;
    char *file;
    int currentLine;
    int currentCol;
    kite_stack_t *obj_created;
    kite_stack_t *decide_exits; /* to fix ticket #17 and short-circuiting */
    char *curStr;
} kite_compiler_t;

/*
 * VM handling functions
 */
kite_vm_t *kite_new_vm(char **args);
void kite_free_vm(kite_vm_t **);

/*
 * Thread handling functions
 */
kite_thread_t *kite_new_thread_without_start(kite_vm_t *);
kite_thread_t *kite_new_thread_bytecode(kite_vm_t *, void *);
kite_thread_t *kite_new_thread_compiled(kite_vm_t *, void *);
#define kite_vm_creator_thread(vm) ((vm)->thread_list[0])
void kite_start_bytecode(kite_thread_t *, void *);
void kite_start_compiled(kite_thread_t *, void *);
 
void kite_join_thread(kite_vm_t *, kite_thread_t *);
void kite_exit_thread(kite_vm_t *, kite_thread_t *);

/*
 * Stack handling functions
 */
void kite_push_stack(kite_stack_t **, int, void *);
void kite_destroy_stack(kite_stack_t **);
void *kite_pop_stack(kite_stack_t **);
void *kite_top_stack(kite_stack_t *);

/*
 * List handling functions
 */
void kite_list_remove(kite_list_t **, kite_list_t **, kite_list_t *);
void kite_list_add_end(kite_list_t **, kite_list_t **, kite_list_t *);
void kite_list_add_begin(kite_list_t **, kite_list_t **, kite_list_t *);

/* 
 * Script operation functions
 */
void *kite_vm_execute(void *);
void kite_vm_execute_user_method(kite_thread_t *, void *);
void kite_vm_execute_exception(kite_thread_t *);
void kite_vm_call_method(kite_thread_t *, struct kite_object_t *, 
                         char *, struct kite_object_t *, int);
void kite_vm_call_object(kite_thread_t *, struct kite_object_t *, 
                         struct kite_object_t *, struct kite_object_t *);
void kite_vm_call_constructor(kite_thread_t *, struct kite_object_t *, 
                              struct kite_object_t *);
int kite_vm_call_operator(kite_thread_t *, struct kite_object_t *, 
                          int, struct kite_object_t *, int);

#define kite_vm_push(thd, obj) \
    kite_push_stack(&(thd)->running_stack, FALSE, kite_reference_object(obj))
#define kite_vm_pop(thd) (struct kite_object_t*)kite_pop_stack(&(thd)->running_stack)
#define kite_vm_return(thd, obj) \
	{ \
        kite_object_t *tmp = (obj); \
		kite_vm_push((thd), tmp); \
        kite_dereference_object(tmp); \
		return; \
	}

#define STACK_MAX 1024

#define PUSH_FUNC_STACK(t) \
     kite_funccall_info_t *entry = malloc(sizeof(kite_funccall_info_t)); \
     kite_funccall_info_t *oldentry = ((kite_funccall_info_t*)kite_top_stack(thd->func_stack)); \
     int old_sys = oldentry ? oldentry->sys_or_user : 0; \
     entry->this = t; \
     entry->locals = NULL; \
     entry->file = ins->common.file; \
     entry->line = ins->common.line; \
     entry->sys_or_user = 0; \
     kite_push_stack(&thd->func_stack, FALSE, entry); \
     if (thd->func_stack->length >= STACK_MAX && !old_sys) { \
         kite_object_t *exc; \
         entry->sys_or_user = 1; \
         exc = kite_new_exception(thd, "System.exceptions.StackOverflow", \
                                  "The function call stack has overflowed."); \
         kite_vm_call_method(thd, exc, "throw", kite_new_list(thd), TRUE); \
         entry->sys_or_user = 0; \
         return; \
     }

#define POP_FUNC_STACK \
    kite_pop_stack(&thd->func_stack); \
    if (entry->locals) kite_destruct_symtab(thd, &entry->locals, TRUE); \
    free(entry);

/*
 * Compiler functions
 */
struct kite_object_t *kite_vm_compile_from_file(kite_thread_t *, char *);
struct kite_object_t *kite_vm_compile_from_fp(kite_thread_t *, FILE *, char *);
struct kite_object_t *kite_vm_compile_from_string(kite_thread_t *, char *);

/*
 * Signal handling.
 */
extern struct kite_object_t *kite_signal_handlers[NSIG + 1];
void kite_handle_signal(int);

/*
 * To facilitate module loading
 */
#ifndef KITE_MODULE_INITIALIZER_C
extern void kite_module_initialize_now(kite_thread_t*);
#endif // KITE_MODULE_INITIALIZER_C

extern inline void *kite_top_stack(kite_stack_t *st)
{
    kite_stack_entry_t *entry;                                                                                   
    if (!st || st->length == 0) return NULL;

    entry = &st->stack[st->length - 1];
    if (entry->reference) {
        return ((kite_symtab_t*)entry->obj)->value;
    }
    return entry->obj;
}

extern inline void *kite_pop_stack(kite_stack_t **st)
{
    kite_stack_t *entry;
    void *ret;

    assert(st != NULL && (*st)->length > 0);
    entry = *st;

    ret = kite_top_stack(entry);
    entry->length--;

    return ret;
}

extern inline void kite_push_stack(kite_stack_t **st, int ref, void *obj)
{
    assert(st != NULL && obj != NULL);

    kite_stack_t *cur = *st;
    if (cur == NULL)
    {
        cur = malloc(sizeof(kite_stack_t));
        assert(cur != NULL);
        cur->length = 0;
        cur->allocated = sizeof(kite_stack_entry_t);
        cur->stack = malloc(cur->allocated);
        *st = cur;
    }

    if (((cur->length + 1) * sizeof(kite_stack_entry_t)) > cur->allocated)
    {
        int newsize = cur->allocated << 1;
        cur->stack = realloc(cur->stack, newsize);
        assert(cur->stack != NULL);
        cur->allocated = newsize;
    }

    kite_stack_entry_t *entry = &cur->stack[++cur->length - 1];
    entry->obj = obj;
    entry->reference = ref;
}

#endif /* KITE_VM__KITEVM_H */
